#ifndef _MODEL_H_ 
#define _MODEL_H_
/*
Com S 476/576 Course Software 
Release 0.1  02/22/00

Copyright 1999,2000

Steven M. LaValle, Iowa State University
*/

class Model {
 protected:
  list<vector> Inputs;
  vector RungeKuttaIntegrate(const vector &x, const vector &u, const double &h);
  vector EulerIntegrate(const vector &x, const vector &u, const double &h);
 public:
  string FilePath;
  double Speed;        // A scalar that can be used to easily adjust
                       //    the step size.
  vector LowerState;   // Vector of minimum values for each state variable
  vector UpperState;   // Vector of maximum values for each state variable
  int StateDim;
  int InputDim;
  Model(string path) {};
  virtual ~Model() {};
  void ReadInputs();
  void ReadLowerState();
  void ReadUpperState();
  virtual list<vector> GetInputs(const vector &x);
  virtual vector StateTransitionEquation(const vector &x, const vector &u) = 0;
  virtual bool Satisfied(const vector &x);
  virtual vector Integrate(const vector &x, const vector &u, 
			   const double &h) = 0; 
  virtual vector LinearInterpolate(const vector &x1, const vector &x2, 
				   const double &a);  // Depends on topology

  // Conversions
  virtual vector StateToConfiguration(const vector &x);
  virtual point  StateToPoint(const vector &x);

  virtual double Metric(const vector &x1, const vector &x2);  // Make a 
              // default metric (it may or may not be used by a planner
};



class ModelLinear: public Model {
 public:
  matrix A;
  matrix B;
  ModelLinear(string path);
  virtual ~ModelLinear() {};
  void ReadA();  // Read state transition matrix
  void ReadB();  // Read input matrix
  virtual point StateToPoint(const vector &x);
  virtual vector StateToConfiguration(const vector &x);
  virtual vector Integrate(const vector &x, const vector &u, const double &h);
  virtual vector StateTransitionEquation(const vector &x, const vector &u);
};



// A planning problem in a 2D world
// There are two derived classes: Model2DPoint and Model2DRigid
class Model2D: public Model {
 public:
  Model2D(string path);
  virtual ~Model2D() {};
};


// A point robot in a 2D world
class Model2DPoint: public Model2D {
 public:
  Model2DPoint(string path);
  virtual ~Model2DPoint() {};
  virtual vector Integrate(const vector &x, const vector &u, const double &h);
  virtual vector StateTransitionEquation(const vector &x, const vector &u);
  virtual double Metric(const vector &x1, const vector &x2);
};


// A point car-like robot in a 2D world
class Model2DPointCar: public Model2DPoint {
 public:
  double MaxSteeringAngle;
  double CarLength;
  Model2DPointCar(string path);
  virtual ~Model2DPointCar() {};
  virtual vector Integrate(const vector &x, const vector &u, const double &h);
  virtual vector StateTransitionEquation(const vector &x, const vector &u);
  virtual double Metric(const vector &x1, const vector &x2);
};


// A rigid robot in a 2D world.  This class can be used for holonomic
// planning in Re^2 X S^1.  Nonholonomic problems are derived.
class Model2DRigid: public Model2D {
 public:
  Model2DRigid(string path);
  virtual ~Model2DRigid() {};
  virtual vector Integrate(const vector &x, const vector &u, const double &h);
  virtual vector StateTransitionEquation(const vector &x, const vector &u);
  vector LinearInterpolate(const vector &x1, const vector &x2, 
					 const double &a);
  virtual double Metric(const vector &x1, const vector &x2);
};

// A rigid car-like robot in a 2D world
class Model2DRigidCar: public Model2DRigid {
 public:
  double MaxSteeringAngle;
  double CarLength;
  Model2DRigidCar(string path);
  virtual ~Model2DRigidCar() {};
  virtual vector StateTransitionEquation(const vector &x, const vector &u);
};


// A rigid car-like robot that can only go forward in a 2D world
class Model2DRigidCarForward: public Model2DRigidCar {
 public:
  Model2DRigidCarForward(string path);
  virtual ~Model2DRigidCarForward() {};
};


// A rigid car-like robot with continuous steering angles
// This model is used by Th. Fraichard, Scheuer, Laugier
class Model2DRigidCarSmooth: public Model2DRigidCar {
 public:
  double SteeringSpeed;
  Model2DRigidCarSmooth(string path);
  virtual ~Model2DRigidCarSmooth() {};
  virtual vector StateTransitionEquation(const vector &x, const vector &u);
  vector Integrate(const vector &x, const vector &u, const double &h);
  virtual double Metric(const vector &x1, const vector &x2);
};



// A 5DOF dynamical model of a rigid car.  This model uses a linear
// tire model, which is far from reality.  The model was donated by
// Jim Bernard (who has made no claims on its accuracy).
class Model2DRigidDyncar: public Model2DRigid {
 public:
  double Mass;       // slugs
  double CAF;        // Front cornering stiffness
  double CAR;        // Rear cornering stiffness
  double Adist;      // Mass center to front tires - feet
  double Bdist;      // Mass center to rear tires - feet
  double Izz;        // yaw moment of interia - ft slugs^2
  double WorldScale; // Feet per world unit (100x100 world)
  double MaxSteeringAngle;
  Model2DRigidDyncar(string path);
  virtual ~Model2DRigidDyncar() {};
  vector Integrate(const vector &x, const vector &u, const double &h);
  virtual point StateToPoint(const vector &x);
  virtual vector StateToConfiguration(const vector &x);
  virtual vector StateTransitionEquation(const vector &x, const vector &u);
  virtual double Metric(const vector &x1, const vector &x2);
};



// A rigid body with two small side thrusters, and a larger lower
// thruster.  The goal is to navigate and softly "land" the craft by
// firing thrusters, in spite of gravity.
class Model2DRigidLander: public Model2DRigid {
 public:
  double Mass;       // kg
  double G;          // Accel of gravity (m/s)
  double Fs;         // Side thruster force
  double Fu;         // Upward thruster force
  Model2DRigidLander(string path);
  virtual ~Model2DRigidLander() {};
  vector Integrate(const vector &x, const vector &u, const double &h);
  virtual vector StateToConfiguration(const vector &x);
  virtual vector StateTransitionEquation(const vector &x, const vector &u);
  virtual double Metric(const vector &x1, const vector &x2);
};


//frode
// A rigid body (helecopter) in a 3D world with 6DOF.
//
class model3DRigid: public Model2D {
public:
  model3DRigid(string path);
  virtual ~model3DRigid() {};

  virtual vector Integrate(const vector &x, const vector &u, const double &h);
  virtual vector StateTransitionEquation(const vector &x, const vector &u);
//  vector LinearInterpolate(const vector &x1, const vector &x2, 
//			   const double &a);
  virtual double Metric(const vector &x1, const vector &x2);
  
};
#endif





