/*
Com S 476/576 Course Software 
Release 0.1  02/22/00

Copyright 1999,2000

Steven M. LaValle, Iowa State University
*/

#include <math.h>
#include <stdio.h>

#include <LEDA/plane_alg.h>
#include <LEDA/file.h>
#include <LEDA/matrix.h>
#include <LEDA/graph.h>
#include <LEDA/list.h>
#include <LEDA/ps_file.h>
#include <LEDA/vector.h>

#ifndef PI
#define PI 3.1415926535897932385
#endif

#ifndef INFINITY
#define INFINITY 1.0e40
#endif


// Stuff that makes the LEDA plane algorithms happy
typedef point   POINT;
typedef segment SEGMENT;
typedef circle  CIRCLE;
typedef line    LINE;
typedef vector  VECTOR;
typedef double  COORD;

#include "world.h"
#include "model.h"
#include "problem.h"
#include "rrt.h"


RRT::RRT(Problem *problem) {
  P = problem;
  NumNodes = 1000;
  MaxNodes = 15000;
  DeltaT = 1.0;  // This is what you get by default
  GoalDist = P->Metric(P->InitialState,P->GoalState);
}



// Return the best new state in nx_best
// success will be false if no action is collision free
vector RRT::SelectBestInput(const vector &x1, const vector &x2, 
			    vector &nx_best, bool &success,
			    bool forward = true)
{
  vector u,u_best,nx;
  double d,d_min;
  success = false;
  d_min = P->Metric(x1,x2); //INFINITY;
  list<vector> il = P->GetInputs(x1);
  forall(u,il) {
    if (forward) 
      nx = P->Integrate(x1,u,DeltaT);
    else
      nx = P->Integrate(x1,u,-DeltaT); 
    d  = P->Metric(nx,x2);
    if ((d < d_min)&&(P->Satisfied(nx))&&(x1 != nx)) 
      {
	d_min = d; u_best = u; nx_best = nx; success = true;
      }
  }
  return u_best;
}



node RRT::NearestNeighbor(const vector &x, const GRAPH<vector,vector> &g) {
  double d,d_min;
  node n,n_best;
  
  n_best = n = g.first_node(); // Keeps the warnings away
  d_min = INFINITY; d = 0.0;

  forall_nodes(n,g) {
    d = P->Metric(g[n],x);
    if (d < d_min) {
      d_min = d; n_best = n; 
    }
  }

  //cout << "n_best: " << n_best << "info: " << g.inf(n_best) << "\n";
  
  return n_best;
}



bool RRT::Extend(const vector &x, 
		 GRAPH<vector,vector> &g,
		 node &nn, bool forward = true) {
  node n_best;
  vector nx,u_best;
  bool success;
  
  n_best = NearestNeighbor(x,g);
  u_best = SelectBestInput(g.inf(n_best),x,nx,success,forward);

  // nx gets next state
  if (success) {   // If a collision-free input was found
    nn = g.new_node(nx); // Make a new node with state nx
    g.new_edge(n_best,nn,u_best);
    //cout << "New node: " << g.inf(nn) << "   " << "u:" << u_best << "\n";
  }

  return success;
}



// This function essentially iterates Extend until it has to stop
// The same action is used for every iteration
bool RRT::Connect(const vector &x, 
		  GRAPH<vector,vector> &g,
		  node &nn, bool forward = true) {
  node nn_prev,n_best;
  vector nx,nx_prev,u_best;
  bool success;
  double d,d_prev;
  
  n_best = NearestNeighbor(x,g);
  u_best = SelectBestInput(g.inf(n_best),x,nx,success,forward); 
           // nx gets next state
  if (success) {   // If a collision-free input was found
    d = P->Metric(nx,x); d_prev = d;
    nx_prev = nx; // Initialize
    nn = n_best;
    while((P->Satisfied(nx))&&(d <= d_prev))
      {
	nx_prev = nx;
	d_prev = d; nn_prev = nn;
	// Uncomment line below to select best action each time
	//u_best = SelectBestInput(g.inf(nn),x,nx,success,forward); 
	if (forward)
	  nx = P->Integrate(nx_prev,u_best,DeltaT);
	else
	  nx = P->Integrate(nx_prev,u_best,-DeltaT);
	d = P->Metric(nx,x);
	// Uncomment the subsequent two lines to
	//   make each intermediate node added
	//nn = g.new_node(nx_prev); // Make a new node
	//g.new_edge(nn_prev,nn,u_best);
      }
    nn = g.new_node(nx_prev); // Make a new node
    g.new_edge(n_best,nn,u_best);
  }

  return success;
}



// Edges of G are input vectors, and vertices are states
// Note that this method also adds to an existing RRT
void RRT::Explore()
{
  int i;
  vector u_best,nx;
  node nn;
  double d;

  // Make the root node of G
  if (G.number_of_nodes() == 0)
    G.new_node(P->InitialState);

  for (i=0; i < NumNodes; i++) {
    if (Extend(ChooseState(),G,nn)) { 
      d = P->Metric(G.inf(nn),P->GoalState);
      if (d < GoalDist)  {  // Decrease if goal closer
	GoalDist = d;
      }
      if (G.number_of_nodes() % 1000 == 0)
	cout << G.number_of_nodes() << " nodes in the RRT.\n";
    }
  }

}



bool RRT::Plan(double gdist, list<node> &path)
{
  int i;
  double d;
  node n,nn,n_goal;
  vector nx,u_best;

  // Make the root node of G
  if (G.number_of_nodes() == 0) 
    n_goal = G.new_node(P->InitialState);

  i = 0;
  n = NearestNeighbor(P->GoalState,G);
  GoalDist = P->Metric(G.inf(n),P->GoalState);

  while ((i < NumNodes)&&(GoalDist > gdist)) {
    if (Extend(ChooseState(),G,nn)) { 
      d = P->Metric(G.inf(nn),P->GoalState);
      if (d < GoalDist) {  // Decrease if goal closer
	GoalDist = d;
	n_goal = nn;
	cout << "GoalDist " << GoalDist << "\n";
      }
    }
    i++;
  }

  // Get the solution path
  if (GoalDist < gdist) {
    cout << "Successful Path Found\n";
    path = PathToLeaf(n_goal,G);
    return true;
  }
  else
    return false;

}




void RRT::WriteRRT(const string &fname)
{
  G.write(P->FilePath + fname);
}



void RRT::ReadRRT(const string &fname)
{
  G.read(P->FilePath + fname);
}




vector RRT::ChooseState() {
  return RandomState();
}


list<node> RRT::PathToLeaf(const node &n, const GRAPH<vector,vector> &g) 
{
   list<node> path;
   node ni;
   int i;

   path.clear();
   path.push(n); 

   i = 0;
   ni = n;
   while ((ni != g.first_node())&&(i < g.number_of_nodes())) {
     path.push(g.source(g.first_in_edge(ni)));
     ni = g.source(g.first_in_edge(ni));
     i++;
   }

   return path;
}


vector RRT::RandomState() {
  int i;
  double r;
  vector rx;

  rx = P->LowerState;
  for (i = 0; i < P->StateDim; i++) {
      R >> r; 
      rx[i] += r * (P->UpperState[i] - P->LowerState[i]);
    }

  return rx;
}




RRTGoalBias::RRTGoalBias(Problem *p):RRT(p) {
  GoalProb = 0.05;
}



vector RRTGoalBias::ChooseState()
{
  double rv;

  R >> rv;
  if (rv > GoalProb)
    return RandomState();
  else
    return P->GoalState;

}




// Build an RRT that tries to go as far as it can for each edge
RRTCon::RRTCon(Problem *p):RRT(p) {
}


void RRTCon::Explore()
{
  int i;
  vector u_best,nx,nx_prev;
  node nn;


  // Make the root node of G
  if (G.number_of_nodes() == 0)
    G.new_node(P->InitialState);

  for (i=0; i < NumNodes; i++) {
    Connect(ChooseState(),G,nn);
  }
}




RRTExtExt::RRTExtExt(Problem *p):RRT(p) {
}


bool RRTExtExt::Plan(double gdist, list<node> &path)
{
  int i;
  vector u_best,nx,nx2;
  node nn,nn2;
  bool connected;
  list<node> path1;

  connected = false;

  if (G.number_of_nodes() == 0)
    G.new_node(P->InitialState);

  if (GG.number_of_nodes() == 0)
    GG.new_node(P->GoalState);

  nn = G.first_node();
  nn2 = GG.first_node();

  i = 0;
  connected = false;
  while ((i < NumNodes) && (!connected)) {
    if (Extend(ChooseState(),G,nn)) {
      if (Extend(G.inf(nn),GG,nn2,false)) {
	i++;
	if (P->Metric(G.inf(nn),G.inf(nn2)) < gdist) {
	  cout << "CONNECTED!!  Nodes: " << G.number_of_nodes()+GG.number_of_nodes() 
	       << "\n";
	  connected = true;
	  path1 = PathToLeaf(nn2,GG); path1.reverse();
	  path = PathToLeaf(nn,G); path.conc(path1);
	}
      }
    }
    i++;

    if ((!connected) && (Extend(ChooseState(),GG,nn,false))) {
      if (Extend(G.inf(nn),G,nn2)) {
	i++;
	if (P->Metric(GG.inf(nn),G.inf(nn2)) < gdist) {
	  cout << "CONNECTED!!  Nodes: " << G.number_of_nodes()+GG.number_of_nodes() << "\n";
	  connected = true;
	  path1 = PathToLeaf(nn,GG); path1.reverse();
	  path = PathToLeaf(nn2,G); path.conc(path1);
	}
      }
    }
  }

  i++;

  if (!connected)
    cout << "Failure to connect after " << G.number_of_nodes()+GG.number_of_nodes() 
	 << " nodes\n";

  return connected;
}




RRTExtCon::RRTExtCon(Problem *p):RRT(p) {
}



bool RRTExtCon::Plan(double gdist, list<node> &path)
{
  int i;
  vector nx,nx_prev;
  node nn,nn2;
  bool connected;
  list<node> path1;

  connected = false;


  if (G.number_of_nodes() == 0)
    G.new_node(P->InitialState);

  if (GG.number_of_nodes() == 0)
    GG.new_node(P->GoalState);

  nn = G.first_node();
  nn2 = GG.first_node();

  i = 0;
  connected = false;
  while ((i < NumNodes) && (!connected)) {
    if (Extend(ChooseState(),G,nn)) { 
      // Update the goal RRT
      if (Connect(G.inf(nn),GG,nn2,false)) {
	if (P->Metric(G.inf(nn),GG.inf(nn2)) < gdist) {
	  cout << "CONNECTED!!  Nodes: " << G.number_of_nodes()+GG.number_of_nodes()
	       << "\n";
	  connected = true;
	  path1 = PathToLeaf(nn2,GG); path1.reverse();
	  path = PathToLeaf(nn,G); path.conc(path1);
	}
      }
    }
    i++;

    if ((!connected)&&(Extend(ChooseState(),GG,nn,false))) { 
      // Update the initial RRT
      if (Connect(GG.inf(nn),G,nn2)) { 
	if (P->Metric(GG.inf(nn),G.inf(nn2)) < gdist) {
	  cout << "CONNECTED!!  Nodes: " << G.number_of_nodes()+GG.number_of_nodes()
	       << "\n";
	  connected = true;
	  path1 = PathToLeaf(nn,GG); path1.reverse();
	  path = PathToLeaf(nn2,G); path.conc(path1);
	}
      }
    }
    i++;
  }

  if (!connected)
    cout << "Failure to connect after " << G.number_of_nodes()+GG.number_of_nodes() 
	 << " nodes\n";

  return connected;
}



RRTConCon::RRTConCon(Problem *p):RRT(p) {
}



bool RRTConCon::Plan(double gdist, list<node> &path)
{
  int i;
  vector nx,nx_prev;
  node nn,nn2;
  bool connected;
  list<node> path1;

  connected = false;

  if (G.number_of_nodes() == 0)
    G.new_node(P->InitialState);

  if (GG.number_of_nodes() == 0)
    GG.new_node(P->GoalState);

  nn = G.first_node();
  nn2 = GG.first_node();

  i = 0;
  connected = false;
  while ((i < NumNodes) && (!connected)) {
    if (Connect(ChooseState(),G,nn)) { 
      // Update the goal RRT
      if (Connect(G.inf(nn),GG,nn2,false)) {
	if (P->Metric(G.inf(nn),GG.inf(nn2)) < gdist) {
	  cout << "CONNECTED!!  Nodes: " << G.number_of_nodes()+GG.number_of_nodes() 
	       << "\n";
	  connected = true;
	  path1 = PathToLeaf(nn2,GG); path1.reverse();
	  path = PathToLeaf(nn,G); path.conc(path1);
	}
      }
    }
    i++;

    if ((!connected)&&(Connect(ChooseState(),GG,nn,false))) { 
      // Update the initial RRT
      if (Connect(GG.inf(nn),G,nn2)) { 
	if (P->Metric(GG.inf(nn),G.inf(nn2)) < gdist) {
	  cout << "CONNECTED!!  Nodes: " << G.number_of_nodes()+GG.number_of_nodes()  
	       << "\n";
	  connected = true;
	  path1 = PathToLeaf(nn,GG); path1.reverse();
	  path = PathToLeaf(nn2,G); path.conc(path1);
	}
      }
    }
    i++;
  }

  if (!connected)
    cout << "Failure to connect after " << G.number_of_nodes()+GG.number_of_nodes() 
	 << " nodes\n";

  return connected;
}

