/*
Com S 476/576 Course Software 
Release 0.1  02/22/00

Copyright 1999,2000

Steven M. LaValle, Iowa State University
*/

#include <fstream.h>
#include <math.h>


#include <LEDA/file.h>
#include <LEDA/array.h>
#include <LEDA/list.h>
#include <LEDA/matrix.h>
#include <LEDA/point.h>
#include <LEDA/polygon.h>
#include <LEDA/stream.h>
#include <LEDA/string.h>
#include <LEDA/vector.h>
#include <LEDA/graph.h>
#include "MatVec.h"

#include "world.h"

#ifndef PI
#define PI 3.1415926535897932385
#endif
#ifndef INFINITY
#define INFINITY 1.0e40
#endif
#ifndef sqr
#define sqr(x) ((x)*(x))
#endif
#ifndef min
#define min(x,y) ((x<y) ? x : y)
#endif

// Some models will interpolate differently because of 
// topology (e.g., S^1, P^3)
// This is the Euclidean default
vector World::LinearInterpolate(const vector &x1, const vector &x2, 
				const double &a) {
  return a*x1 + (1.0-a)*x2;
}


// *********************************************************************
// *********************************************************************
// CLASS:     World::WorldPolygonal
// 
// *********************************************************************
// *********************************************************************

// Constructor
WorldPolygonal::WorldPolygonal(string path = ""):World(path) {

  if ((path.length() > 0)&&(path[path.length()-1] != '/'))
    path += "/";

  FilePath = path;
  
  if (is_file(FilePath + "Obst")) {
    file_istream fin(FilePath + "Obst");
    fin >> Obst;
  }
}



// This simple collision checker uses LEDA's p.inside method.
// Certainly more efficient methods exist (incremental).
bool WorldPolygonal::CollisionFree(const vector &x) {
  polygon p;

  forall(p,Obst) {
    if (p.inside(point(x))) {
      return false;
    }
  }

  return true;
}


double WorldPolygonal::DistanceComp(const vector &x) {
  polygon p;
  double d_min,d;

  d_min = INFINITY;

  if (CollisionFree(x)) {
    forall(p,Obst) {
      d = p.distance(point(x));
      if (d < d_min)
	d_min = d;
    }
  }
  else
    d_min = -1.0;
  
  return d_min;
}



list<polygon> WorldPolygonal::EnvironmentToPolygons() {
  return Obst;
}


list<polygon> WorldPolygonal::RobotToPolygons(const vector &x) {
  list<polygon> trobot;

  point p1(0.5,0.5);
  point p2(-0.5,0.5);
  point p3(-0.5,-0.5);
  point p4(0.5,-0.5);
  list<point> pl;
  pl.push(p4); pl.push(p3); pl.push(p2); pl.push(p1);
  polygon p(pl);
  trobot.push(p.translate(x));
  
  return trobot;
}



// *********************************************************************
// *********************************************************************
// CLASS:     World::WorldPolygonalRigid
// 
// *********************************************************************
// *********************************************************************

// Constructor
WorldPolygonalRigid::WorldPolygonalRigid(string path = ""):WorldPolygonal(path) {

  if ((path.length() > 0)&&(path[path.length()-1] != '/'))
    path += "/";

  FilePath = path;
  
  if (is_file(FilePath + "Robot")) {
    file_istream fin(FilePath + "Robot");
    fin >> Robot;
  }
}


// This is a very naive collision checker
// This should certainly be replaced some day with a 
// a more efficient algorithm.  It would be best to add
// a 2D incremental distance algorithm, based on the
// ideas of Lin and Canny, and Mirtich.  One could also
// use bitmaps since this is only 2D.

bool WorldPolygonalRigid::CollisionFree(const vector &x) {
  polygon op,rp;
  segment seg,seg2;

  list<polygon> trobot = RobotToPolygons(x);

  forall(rp,trobot) {
    forall(op,Obst) {
      forall_segments(seg,rp) {
	forall_segments(seg2,op) {
	  if (seg.intersection(seg2)) {
	    return false;
	  }
	}
      }
    }
  }

  // Check for containment
  forall(rp,trobot) {
    forall(op,Obst) {
      if (op.contains(rp.vertices().head()))
	return false;
    }
  }

  return true;
}


// This should certainly be replaced some day with a 
// a more efficient algorithm.  It would be best to add
// a 2D incremental distance algorithm, based on the
// ideas of Lin and Canny, and Mirtich.

double WorldPolygonalRigid::DistanceComp(const vector &x) {
  polygon rp,op;
  double d_min,d;
  point rpt;

  d_min = INFINITY;

  list<polygon> trobot = RobotToPolygons(x);

  // This is obviously inefficient
  if (CollisionFree(x)) {
    forall(rp,trobot) {
      forall_vertices(rpt,rp) {
	forall(op,Obst) {
	  d = op.distance(rpt);
	  if (d < d_min)
	    d_min = d;
	}
      }
    }
  }
  else
    d_min = -1.0;
  
  return d_min;
}


// Handle S^1 topology properly (for rotation)
vector WorldPolygonalRigid::LinearInterpolate(const vector &x1, const vector &x2, 
					 const double &a) {

  vector v;

  v = a*x1 + (1.0-a)*x2;

  if (x2[2] - x1[2] > PI) {
    if (x1[2] > x2[2])
      v[2] = a*x1[2] + (1.0-a)*(x2[2]+PI);
    else
      v[2] = a*(x1[2]+PI) + (1.0-a)*x2[2];
  }

  if (v[2] > 2.0*PI)
    v[2] -= 2.0*PI;

  return v;
}



list<polygon> WorldPolygonalRigid::RobotToPolygons(const vector &q) {
  list<polygon> trobot;
  polygon p,tp;

  forall(p,Robot) {
    tp = p.rotate(q[2]).translate(q[0],q[1]);
    trobot.push_front(tp);
  }

  return trobot;
}


// *********************************************************************
// *********************************************************************
// CLASS:     World::worldPQP
// BY   :     Frode Aarstad 
// *********************************************************************
// *********************************************************************


worldPQP::worldPQP(maze3D *m, robot *r, string path = "")
  :World(path) 
{  
  theMaze = m;
  theRobot = r;
}


worldPQP::~worldPQP()
{
  delete theMaze;
  delete theRobot;
}



bool worldPQP::CollisionFree(const vector &x) {
  GLdouble myMat[16];

  glPushMatrix();
    if (x.dim() == 2) 
      glTranslatef(x[0], 0.2, x[1]);
    else {
      glTranslatef(x[0], x[1], x[2]);
      glRotatef((x[3]*360)/2.0*PI, 0.0, 1.0, 0.0);
    }
    glGetDoublev(GL_MODELVIEW_MATRIX, myMat);
    OGLtoMV(theRobot->R, theRobot->T, myMat);
  glPopMatrix();

  return theMaze->collisionFree(theRobot, 0.5);
}


double worldPQP::DistanceComp(const vector &x) {
  GLdouble myMat[16];
  
  glPushMatrix();
  if (x.dim() == 2) 
    glTranslatef(x[0], 0.2, x[1]);
  else {
    glTranslatef(x[0], x[1], x[2]);
    glRotatef((x[3]*360)/2.0*PI, 0.0, 1.0, 0.0);
  }
  glGetDoublev(GL_MODELVIEW_MATRIX, myMat);
  OGLtoMV(theRobot->R, theRobot->T, myMat);
  glPopMatrix();
  
  return (double)theMaze->collisionDistance(theRobot);
}



vector worldPQP::LinearInterpolate(const vector &q1, const vector &q2, 
                                   const double &a) 
{
  vector vec;
  
  return vec;
}



list<polygon> worldPQP::EnvironmentToPolygons() 
{
  list<polygon> p;
  return p;
}


list<polygon> worldPQP::RobotToPolygons(const vector &q) 
{
  list<polygon> p;
  return p;
}





